@@ -17,15 +17,22 @@
 //!   - [`{% extends %}`](#-extends-)
 //!   - [`{% block %}`](#-block-)
 //!   - [`{% include %}`](#-include-)
+//!   - [`{% import %}`](#-import-)
 //!   - [`{% with %}`](#-with-)
+//!   - [`{% set %}`](#-set-)
 //!   - [`{% filter %}`](#-filter-)
+//!   - [`{% macro %}`](#-macro-)
+//!   - [`{% call %}`](#-call-)
+//!   - [`{% do %}`](#-do-)
 //!   - [`{% autoescape %}`](#-autoescape-)
+//!   - [`{% raw %}`](#-raw-)
+//! - [Custom Delimiters](#custom-delimiters)
 //!
 //! </details>
 //!
 //! # Synopsis
 //!
-//! A MiniJinja template is simply a text file.  MiniJInja can generate any text-based
+//! A MiniJinja template is simply a text file.  MiniJinja can generate any text-based
 //! format (HTML, XML, CSV, LaTeX, etc.).  A template doesn’t need to have a specific extension
 //! and in fact MiniJinja does not understand much about the file system.  However the default
 //! configuration for [auto escaping](crate::Environment::set_auto_escape_callback) uses file
@@ -67,7 +74,7 @@
 //!
 //! ## Literals
 //!
-//! The simplest form of expressions are literals. Literals are representations for Python
+//! The simplest form of expressions are literals. Literals are representations for
 //! objects such as strings and numbers. The following literals exist:
 //!
 //! - `"Hello World"`: Everything between two double or single quotes is a string. They are
@@ -76,8 +83,8 @@
 //! - `42`: Integers are whole numbers without a decimal part.
 //! - `42.0`: Floating point numbers can be written using a `.` as a decimal mark.
 //! - `['list', 'of', 'objects']`: Everything between two brackets is a list. Lists are useful
-//!   for compatibility with Jinja2 `('list', 'of', 'objects')` is also allowed.
 //!   for storing sequential data to be iterated over.
+//!   for compatibility with Jinja2 `('list', 'of', 'objects')` is also allowed.
 //! - `{'map': 'of', 'key': 'and', 'value': 'pairs'}`: A map is a structure that combines keys
 //!   and values. Keys must be unique and always have exactly one value. Maps are rarely
 //!   created in templates.
@@ -91,7 +98,7 @@
 //! - ``+``: Adds two numbers up. ``{{ 1 + 1 }}`` is ``2``.
 //! - ``-``: Subtract the second number from the first one.  ``{{ 3 - 2 }}`` is ``1``.
 //! - ``/``: Divide two numbers. ``{{ 1 / 2 }}`` is ``0.5``.  See note on divisions below.
-//! - ``//``: Integer divide two numbers. ``{{ 5 // 3 }}`` is ``1``.  See note on divisons below.
+//! - ``//``: Integer divide two numbers. ``{{ 5 // 3 }}`` is ``1``.  See note on divisions below.
 //! - ``%``: Calculate the remainder of an integer division.  ``{{ 11 % 7 }}`` is ``4``.
 //! - ``*``: Multiply the left operand with the right one.  ``{{ 2 * 2 }}`` would return ``4``.
 //! - ``**``: Raise the left operand to the power of the right operand.  ``{{ 2**3 }}``
@@ -106,8 +113,8 @@
 //! - ``!=``: Compares two objects for inequality.
 //! - ``>``: ``true`` if the left hand side is greater than the right hand side.
 //! - ``>=``: ``true`` if the left hand side is greater or equal to the right hand side.
-//! - ``<``:``true`` if the left hand side is lower than the right hand side.
-//! - ``<=``: ``true`` if the left hand side is lower or equal to the right hand side.
+//! - ``<``:``true`` if the left hand side is less than the right hand side.
+//! - ``<=``: ``true`` if the left hand side is less or equal to the right hand side.
 //!
 //! ## Logic
 //!
@@ -133,7 +140,13 @@
 //!   can use positional arguments.  Additionally keyword arguments are supported
 //!   which are treated like a dict syntax.  Eg: `foo(a=1, b=2)` is the same as
 //!   `foo({"a": 1, "b": 2})`.
-//! - ``.`` / ``[]``: Get an attribute of an object.
+//! - ``.`` / ``[]``: Get an attribute of an object.  If an object does not have a specific
+//!   attribute or item then `undefined` is returned.  Accessing a property of an already
+//!   undefined value will result in an error.
+//! - ``[start:stop]`` / ``[start:stop:step]``: slices a list or string.  All three expressions
+//!   are optional (`start`, `stop`, `step`).  For instance ``"Hello World"[:5]`` will return
+//!   just `"Hello"`.  Likewise ``"Hello"[1:-1]`` will return `"ell"`.  The step component can
+//!   be used to change the step size.  `"12345"[::2]` will return `"135"`.
 //!
 //! ### If Expressions
 //!
@@ -193,9 +206,13 @@
 //! - `loop.cycle`: A helper function to cycle between a list of sequences. See the explanation below.
 //! - `loop.depth`: Indicates how deep in a recursive loop the rendering currently is. Starts at level 1
 //! - `loop.depth0`: Indicates how deep in a recursive loop the rendering currently is. Starts at level 0
+//! - `loop.previtem`: The item from the previous iteration of the loop. `Undefined` during the first iteration.
+//! - `loop.nextitem`: The item from the previous iteration of the loop. `Undefined` during the last iteration.
+//! - `loop.changed(...args)`: Returns true if the passed values have changed since the last time it was called with the same arguments.
+//! - `loop.cycle(...args)`: Returns a value from the passed sequence in a cycle.
 //!
 //! Within a for-loop, it’s possible to cycle among a list of strings/variables each time through
-//! the loop by using the special loop.cycle helper:
+//! the loop by using the special `loop.cycle` helper:
 //!
 //! ```jinja
 //! {% for row in rows %}
@@ -203,6 +220,19 @@
 //! {% endfor %}
 //! ```
 //!
+//! A `loop.changed()` helper is also available which can be used to detect when
+//! a value changes between the last iteration and the current one.  The method
+//! takes one or more arguments that are all compared.
+//!
+//! ```jinja
+//! {% for entry in entries %}
+//!   {% if loop.changed(entry.category) %}
+//!     <h2>{{ entry.category }}</h2>
+//!   {% endif %}
+//!   <p>{{ entry.message }}</p>
+//! {% endfor %}
+//! ```
+//!
 //! Unlike in Rust or Python, it’s not possible to break or continue in a loop. You can,
 //! however, filter the sequence during iteration, which allows you to skip items.  The
 //! following example skips all the users which are hidden:
@@ -243,6 +273,10 @@
 //! </ul>
 //! ```
 //!
+//! **Special note:** the `previtem` and `nextitem` attributes are available by default
+//! but can be disabled by removing the `adjacent_loop_items` crate feature.  Removing
+//! these attributes can provide meaningful speedups for templates with a lot of loops.
+//!
 //! ## `{% if %}`
 //!
 //! The `if` statement is comparable with the Python if statement. In the simplest form,
@@ -273,6 +307,8 @@
 //!
 //! ## `{% extends %}`
 //!
+//! **Feature:** `multi_template` (included by default)
+//!
 //! The `extends` tag can be used to extend one template from another.  You can have multiple
 //! `extends` tags in a file, but only one of them may be executed at a time.  For more
 //! information see [block](#-block-).
@@ -354,10 +390,12 @@
 //!
 //! However, the name after the `endblock` word must match the block name.
 //!
-//! ## `{% include #}`
+//! ## `{% include %}`
+//!
+//! **Feature:** `multi_template` (included by default)
 //!  
 //! The `include` tag is useful to include a template and return the rendered contents of that file
-//! into the current namespace::
+//! into the current namespace:
 //!  
 //! ```jinja
 //! {% include 'header.html' %}
@@ -384,6 +422,35 @@
 //!  
 //! Included templates have access to the variables of the active context.
 //!
+//! ## `{% import %}`
+//!
+//! **Feature:** `multi_template` (included by default)
+//!
+//! MiniJinja supports the `{% import %}` and `{% from ... import ... %}`
+//! syntax.  With it variables or macros can be included from other templates:
+//!
+//! ```jinja
+//! {% from "my_template.html" import my_macro, my_variable %}
+//! ```
+//!
+//! Imports can also be aliased:
+//!
+//! ```jinja
+//! {% from "my_template.html" import my_macro as other_name %}
+//! {{ other_name() }}
+//! ```
+//!
+//! Full modules can be imported with `{% import ... as ... %}`:
+//!
+//! ```jinja
+//! {% import "my_template.html" as helpers %}
+//! {{ helpers.my_macro() }}
+//! ```
+//!
+//! Note that unlike Jinja2, exported modules do not contain any template code.  Only
+//! variables and macros that are defined can be imported.  Also imports unlike in Jinja2
+//! are not cached and they get access to the full template context.
+//!
 //! ## `{% with %}`
 //!
 //! The with statement makes it possible to create a new inner scope.  Variables set within
@@ -396,6 +463,47 @@
 //! foo is not visible here any longer
 //! ```
 //!
+//! Multiple variables can be set at once and unpacking is supported:
+//!
+//! ```jinja
+//! {% with a = 1, (b, c) = [2, 3] %}
+//!   {{ a }}, {{ b }}, {{ c }}  (outputs 1, 2, 3)
+//! {% endwith %}
+//! ```
+//!
+//! ## `{% set %}`
+//!
+//! The `set` statement can be used to assign to variables on the same scope.  This is
+//! similar to how `with` works but it won't introduce a new scope.
+//!
+//! ```jinja
+//! {% set navigation = [('index.html', 'Index'), ('about.html', 'About')] %}
+//! ```
+//!
+//! Please keep in mind that it is not possible to set variables inside a block
+//! and have them show up outside of it.  This also applies to loops.  The only
+//! exception to that rule are if statements which do not introduce a scope.
+//!
+//! It's also possible to capture blocks of template code into a variable by using
+//! the `set` statement as a block.   In that case, instead of using an equals sign
+//! and a value, you just write the variable name and then everything until
+//! `{% endset %}` is captured.
+//!
+//! ```jinja
+//! {% set navigation %}
+//! <li><a href="/">Index</a>
+//! <li><a href="/downloads">Downloads</a>
+//! {% endset %}
+//! ```
+//!
+//! The `navigation` variable then contains the navigation HTML source.
+//!
+//! This can also be combined with applying a filter:
+//!
+//! ```jinja
+//! {% set title | upper %}Title of the page{% endset %}
+//! ```
+//!
 //! ## `{% filter %}`
 //!
 //! Filter sections allow you to apply regular [filters](crate::filters) on a
@@ -407,6 +515,135 @@
 //! {% endfilter %}
 //! ```
 //!
+//! ## `{% macro %}`
+//!
+//! **Feature:** `macros` (included by default)
+//!
+//! MiniJinja has limited support for macros.  They allow you to write reusable
+//! template functions.  They are useful to put often used idioms into reusable
+//! functions so you don't repeat yourself (“DRY”).
+//!
+//! Here’s a small example of a macro that renders a form element:
+//!
+//! ```jinja
+//! {% macro input(name, value="", type="text") -%}
+//! <input type="{{ type }}" name="{{ name }}" value="{{ value }}">
+//! {%- endmacro %}
+//! ```
+//!
+//! The macro can then be called like a function in the namespace:
+//!
+//! ```jinja
+//! <p>{{ input('username') }}</p>
+//! <p>{{ input('password', type='password') }}</p>
+//! ```
+//!
+//! The behavior of macros with regards to undefined variables is that they capture
+//! them at macro declaration time (eg: they use a closure).
+//!
+//! Macros can be imported via `{% import %}` or `{% from ... import %}`.
+//!
+//! Macros also accept a hidden `caller` keyword argument for the use with
+//! `{% call %}`.
+//!
+//! ## `{% call %}`
+//!
+//! **Feature:** `macros` (included by default)
+//!
+//! This tag functions similar to a macro that is passed to another macro.  You can
+//! think of it as a way to declare an anonymous macro and pass it to another macro
+//! with the `caller` keyword argument.  The following example shows a macro that
+//! takes advantage of the call functionality and how it can be used:
+//!
+//! ```jinja
+//! {% macro dialog(title) %}
+//!   <div class="dialog">
+//!     <h3>{{ title }}</h3>
+//!     <div class="contents">{{ caller() }}</div>
+//!   </div>
+//! {% endmacro %}
+//!
+//! {% call dialog(title="Hello World") %}
+//!   This is the dialog body.
+//! {% endcall %}
+//! ```
+//!
+//! <details><summary><strong style="cursor: pointer">Macro Alternative:</strong></summary>
+//!
+//! The above example is more or less equivalent with the following:
+//!
+//! ```jinja
+//! {% macro dialog(title) %}
+//!   <div class="dialog">
+//!     <h3>{{ title }}</h3>
+//!     <div class="contents">{{ caller() }}</div>
+//!   </div>
+//! {% endmacro %}
+//!
+//! {% macro caller() %}
+//!   This is the dialog body.
+//! {% endmacro %}
+//! {{ dialog(title="Hello World", caller=caller) }}
+//! ```
+//!
+//! </details>
+//!
+//! It’s also possible to pass arguments back to the call block.  This makes it
+//! useful to build macros that behave like if statements or loops.  Arguments
+//! are placed surrounded in parentheses right after the `call` keyword and
+//! is followed by the macro to be called.
+//!
+//! ```jinja
+//! {% macro render_user_list(users) %}
+//! <ul>
+//! {% for user in users %}
+//!   <li><p>{{ user.username }}</p>{{ caller(user) }}</li>
+//! {% endfor %}
+//! </ul>
+//! {% endmacro %}
+//!
+//! {% call(user) render_user_list(list_of_user) %}
+//! <dl>
+//!   <dt>Name</dt>
+//!   <dd>{{ user.name }}</dd>
+//!   <dt>E-Mail</dt>
+//!   <dd>{{ user.email }}</dd>
+//! </dl>
+//! {% endcall %}
+//! ```
+//!
+//! ## `{% do %}`
+//!
+//! The do tag has the same functionality as regular template tags (`{{ ... }}`);
+//! except it doesn't output anything when called.
+//!
+//! This is useful if you have a function or macro that has a side-effect, and
+//! you don’t want to display output in the template. The following example
+//! shows a macro that uses the do functionality, and how it can be used:
+//!
+//! ```jinja
+//! {% macro dialog(title) %}
+//!   Dialog is {{ title }}
+//! {% endmacro %}
+//!
+//! {% do dialog(title="Hello World") %} <- does not output anything
+//! ```
+//!
+//! The above example will not output anything when using the `do` tag.
+//!
+//! This tag exists for consistency with Jinja2 and can be useful if you have
+//! custom functionality in templates that uses side-effects.  For instance if
+//! you expose a function to templates that can be used to log warnings:
+//!
+//! ```jinja
+//! {% for user in users %}
+//!   {% if user.deleted %}
+//!     {% log warn("Found unexpected deleted user in template") %}
+//!   {% endif %}
+//!   ...
+//! {% endfor %}
+//! ```
+//!
 //! ## `{% autoescape %}`
 //!
 //! If you want you can activate and deactivate the autoescaping from within
@@ -425,5 +662,61 @@
 //! ```
 //!
 //! After an `endautoescape` the behavior is reverted to what it was before.
+//!
+//! The exact auto escaping behavior is determined by the value of
+//! [`AutoEscape`](crate::AutoEscape) set to the template.
+//!
+//! ## `{% raw %}`
+//!
+//! A raw block is a special construct that lets you ignore the embedded template
+//! syntax.  This is particularly useful if a segment of template code would
+//! otherwise require constant escaping with things like `{{ "{{" }}`:
+//!
+//! Example:
+//!
+//! ```jinja
+//! {% raw %}
+//! <ul>
+//! {% for item in seq %}
+//!     <li>{{ item }}</li>
+//! {% endfor %}
+//! </ul>
+//! {% endraw %}
+//! ```
+//!
+#![cfg_attr(
+    feature = "custom_syntax",
+    doc = r#"
+# Custom Delimiters
 
-// this is just for docs
+When MiniJinja has been compiled with the `custom_syntax` feature (see
+[`Syntax`](crate::Syntax)), it's possible to reconfigure the delimiters of the
+templates.  This is generally not recommended but it's useful for situations
+where Jinja templates are used to generate files with a syntax that would be
+conflicting for Jinja.  With custom delimiters it can for instance be more
+convenient to generate LaTeX files:
+
+```
+# use minijinja::{Environment, Syntax};
+let mut environment = Environment::new();
+environment.set_syntax(Syntax {
+    block_start: "\\BLOCK{".into(),
+    block_end: "}".into(),
+    variable_start: "\\VAR{".into(),
+    variable_end: "}".into(),
+    comment_start: "\\#{".into(),
+    comment_end: "}".into(),
+}).unwrap();
+```
+
+And then a template might look like this instead:
+
+```latex
+\begin{itemize}
+\BLOCK{for item in sequence}
+  \item \VAR{item}
+\BLOCK{endfor}
+\end{itemize}
+```
+"#
+)]
